# -*- coding: utf-8 -*-
# @Author  : ZillyRex


import os
from multiprocessing import Pool, cpu_count
import cv2.cv2 as cv2


def resize_img(resize_func, path_img, path_img_out,
               path_anno=None, path_anno_out=None,
               path_label=None, path_label_out=None, verbose=0):
    """
    Resize an image and corresponding annotation and/or label file.

    Args:
        resize_func: A function defining the resize strategy which calls two
        parameters, width and height, and returns the new width and height after resizing.
        eg. resize_func(x, y): return (x/2 if x > 2000 else x, y/3 if y > 3000 else y)
        path_img: Path of the image you wanna resize.
        path_img_out: Path of the folder you wanna save the image after resizing.
        path_anno: Path of the annotation file corresponding to the image.
        path_anno_out: Path of the folder you wanna save the annotation file after resizing.
        path_label: Path of the label file corresponding to the image.
        path_label_out: Path of the folder you wanna save the label file after resizing.

    Returns:
        None
    """
    img = cv2.imread(path_img)
    if not img:
        print('Something wrong with the image {}.'.format(path_img))
        return
    H, W, _ = img.shape
    img = cv2.resize(img, (resize_func(W, H)), interpolation=cv2.INTER_NEAREST)

    if not os.path.isdir(path_img_out):
        os.mkdir(path_img_out)
    basename = os.path.basename(path_img)
    cv2.imwrite(os.path.join(path_img_out, basename), img)

    if path_anno:
        if path_anno_out:
            pass
            # TODO: change anno and save it.
        else:
            print('You need to assign the path_anno_out.')
    else:
        print('No path_anno assigned.')

    if path_label:
        if path_label_out:
            pass
            # TODO: change label and save it.
        else:
            print('You need to assign the path_label_out.')
    else:
        print('No path_label assigned.')

    if verbose:
        print(basename)


def resize_imgs(resize_func, path_img_folder, path_img_out,
                path_anno_folder=None, path_anno_out=None,
                path_label_folder=None, path_label_out=None, verbose=0):
    """
    Resize an image and corresponding annotation and/or label file.

    Args:
        resize_func: A function defining the resize strategy which calls two
        parameters, width and height, and returns the new width and height after resizing.
        eg. resize_func(x, y): return (x/2 if x > 2000 else x, y/3 if y > 3000 else y)
        path_img_folder: Path of the folder containing all the images you wanna resize.
        path_img_out: Path of the folder you wanna save the images after resizing.
        path_anno: Path of the annotation files corresponding to the images.
        path_anno_out: Path of the folder you wanna save the annotation files after resizing.
        path_label_folder: Path of the label files corresponding to the images.
        path_label_out: Path of the folder you wanna save the label files after resizing.

    Returns:
        None
    """
    path_imgs = [os.path.join(path_img_folder, i)
                 for i in os.listdir(path_img_folder)]
    resize_func_ = [resize_func for i in range(len(path_imgs))]
    path_img_out_ = [path_img_out for i in range(len(path_imgs))]

    if path_anno_folder:
        path_annos = [os.path.join(path_anno_folder, i)
                      for i in os.listdir(path_anno_folder)]
    else:
        path_annos = [None for i in range(len(path_imgs))]

    path_anno_out_ = [path_anno_out
                      if path_anno_out else path_anno_out
                      for i in range(len(path_imgs))]

    if path_label_folder:
        path_labels = [os.path.join(path_label_folder, i)
                       for i in os.listdir(path_label_folder)]
    else:
        path_labels = [None for i in range(len(path_imgs))]

    path_label_out_ = [path_label_out
                       if path_label_out else path_label_out
                       for i in range(len(path_imgs))]

    verbose_ = [verbose for i in range(len(path_imgs))]

    pool = Pool(cpu_count())
    pool.starmap(resize_img, zip(resize_func_, path_imgs, path_img_out_,
                                 path_annos, path_anno_out_,
                                 path_labels, path_label_out_, verbose_))
    pool.close()
    pool.join()



# -*- coding: utf-8 -*-
# @Author  : ZillyRex


import os
from multiprocessing import Pool, cpu_count
import cv2.cv2 as cv2
from . import analysis


def draw_bbox(img, *args):
    """
    Draw the bounding box on an image.

    Args:
        img: An image instance obtained by cv2.imread.
        args: {cls_, x, y, w, h, color_bbox, color_text}.
            cls_ is object class name.
            x, y are the coordinates of the bbox top-left corner.
            w, h are the weight and height of the bbox.
            color_bbox is the (B, G, R) color for bbox.
            color_bbox is the (B, G, R) color for text.

    Returns:
        The img with bbox on it.
    """
    cls_, x, y, w, h, color_bbox, color_text = args
    cv2.rectangle(img, (x, y), (x+w, y+h), color_bbox, 4)
    cv2.putText(img, cls_, (x, y-10),
                cv2.FONT_HERSHEY_SIMPLEX, 1, color_text, 1)
    return img


def draw_anno(path_img, path_anno, path_out, verbose=0):
    """
    Draw the bounding box of on an image by its annotation file.

    Args:
        path_img: Path of the image.
        path_anno: Path of the corresponding annotation file.
        path_out: Path of the output folder.

    Returns:
        None
    """
    img = cv2.imread(path_img)
    anno = analysis.parse_anno(path_anno)
    color_bbox = (0, 255, 0)
    color_text = (0, 0, 255)
    params = []
    for obj in anno['object']:
        params.append((obj['name'],
                       obj['xmin'], obj['ymin'],
                       obj['xmax']-obj['xmin'], obj['ymax']-obj['ymin'],
                       color_bbox, color_text))
    for param in params:
        draw_bbox(img, *param)
    cv2.imwrite(os.path.join(path_out, '{}'.format(
        os.path.basename(path_img))), img)
    if verbose:
        print(path_img)


def draw_annos(path_img_folder, path_anno_folder, path_out, verbose=0):
    """
    Draw the bounding box of on a list of images by their annotation files.

    Args:
        path_img_folder: Path of the images folder.
        path_anno_folder: Path of the corresponding annotation folder.
        path_out: Path of the output folder.

    Returns:
        None
    """
    if not analysis.is_match(path_anno_folder, path_img_folder):
        print('file names in the images and annotation folders don\'t match each other!')
        return
    if not os.path.isdir(path_out):
        os.mkdir(path_out)
    path_imgs = [os.path.join(path_img_folder, i)
                 for i in os.listdir(path_img_folder)]
    path_annos = [os.path.join(path_anno_folder, i)
                  for i in os.listdir(path_anno_folder)]
    path_out = [path_out for i in range(len(path_imgs))]
    verbose_ = [verbose for i in range(len(path_imgs))]
    pool = Pool(cpu_count())
    pool.starmap(draw_anno, zip(path_imgs, path_annos, path_out, verbose_))
    pool.close()
    pool.join()


def draw_result(path_result, path_img_folder, path_out, verbose=0):
    """
    Draw the bounding box by a result file.
    The result file format is as follow:
        ...... /xxx/xxx/xxx.jpg: ......
        ...... /xxx/xxx/xxx.jpg: ......
        class_name: 75%	(left_x:  626   top_y:  961   width:   50   height:   18)
        class_name: 84%	(left_x:  626   top_y:  943   width:   47   height:   20)

    Args:
        path_result: Path of the result file.
        path_img_folder: Path of the folder containing all the images in the result file.
        path_out: Path of the output folder.

    Returns:
        None
    """
    if not os.path.isdir(path_out):
        os.mkdir(path_out)
    args_list = []
    basename = None
    with open(path_result) as f:
        for line in f:
            if '/' in line:
                if args_list:
                    img = cv2.imread(os.path.join(path_img_folder, basename))
                    for args in args_list:
                        draw_bbox(img, *args)
                    cv2.imwrite(os.path.join(path_out, basename), img)
                    args_list.clear()
                    if verbose:
                        print(basename)
                abspth = line.split()[3][:-1]
                basename = os.path.basename(abspth)
            if '%' in line:
                l = line.strip('\n').split()
                cls_ = l[0][:-1]
                x = int(float(l[3]))
                y = int(float(l[5]))
                w = int(float(l[7]))
                h = int(float(l[9][:-1]))
                args_list.append(
                    (cls_, x, y, w, h, (0, 255, 255), (0, 255, 255)))



import setuptools

with open("README.md", "r", encoding='utf-8') as fh:
    long_description = fh.read()

setuptools.setup(
    name="odutil",
    version="0.0.16",
    author="ZillyRex",
    author_email="zillyrain@gmail.com",
    description="A group of utils for object detection.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/ZillyRex/odutil",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
    ],
)



# -*- coding: utf-8 -*-
# @Author  : ZillyRex


import os
import numpy as np
import xml.etree.ElementTree as ET
from multiprocessing import Pool, cpu_count


def parse_anno(path_anno):
    """
    Parse an annotation file into a dict.

    Args:
        path_anno: The path of the annotation file you wanna parse.

    Returns:
        A dict mapping filename, size and objects in the annotation to the
        corresponding data fetched.
        For example:

        {'annoname': 'xxx.xml',
         'filename': 'image.jpg',
         'size': {'width': '1621', 'height': '1216', 'depth': '3'},
         'object': [
             {'name': 'class1', 'xmin': '904', 'ymin': '674', 'xmax': '926', 'ymax': '695'},
             {'name': 'class2', 'xmin': '972', 'ymin': '693', 'xmax': '993', 'ymax': '713'}]}
    """
    res = {}
    tree = ET.ElementTree(file=path_anno)

    # Parse annotation name
    res['annoname'] = os.path.basename(path_anno)

    # Parse filename
    res['filename'] = tree.find('filename').text

    # Parse size
    size = tree.find('size')
    dict_size = {}
    for item in size:
        dict_size[item.tag] = int(float(item.text))
    res['size'] = dict_size

    # Parse object
    objs = tree.findall('object')
    res['object'] = []
    for obj in objs:
        dict_obj = {}
        dict_obj['name'] = obj.find('name').text
        bbox = obj.find('bndbox')
        for item in bbox:
            dict_obj[item.tag] = int(float(item.text))
        res['object'].append(dict_obj)
    return res


def parse_annos(path_anno_folder):
    """
    Parse a list of annotation files into a list of dicts.

    Args:
        path_anno_folder: Path of the directory of the annotation files you wanna parse.

    Returns:
        A dict of dicts. Each of them mapping annotation file name("annoname")
        to the corresponding annotation dict fetched by parse_anno().
    """
    path_annos = [os.path.join(path_anno_folder, i)
                  for i in os.listdir(path_anno_folder)]
    pool = Pool(cpu_count())
    res = pool.map(parse_anno, path_annos)
    pool.close()
    pool.join()
    annos_dict = {}
    for anno in res:
        annos_dict[os.path.splitext(anno['annoname'])[0]] = anno
    return annos_dict


def is_match(path_1, path_2):
    """
    Check if the file names in the folders match each other.

    Param:
        path_1: Path of a folder.
        path_2: Path of another folder.

    Return:
        True if match else False.
    """
    name_1 = os.listdir(path_1)
    name_2 = os.listdir(path_2)
    if len(name_1) != len(name_2):
        return False
    set_name = set()
    for name in name_1:
        set_name.add(os.path.splitext(name)[0])
    for name in name_2:
        if os.path.splitext(name)[0] not in set_name:
            return False
    return True


def anno2label(path_anno, path_names, path_out):
    """
    Generate label txt from annotation.

    Args:
        path_anno: Path of the annotation file.
        path_names: Path of the .names file. Only the class in the .names will be converted.
        path_out: Path of the output .txt file.

    Returns:
        None
    """
    anno = parse_anno(path_anno)
    name2label, _ = get_label(path_names)

    W, H = anno['size']['width'], anno['size']['height']
    row = []
    for bbox in anno['object']:
        if bbox['name'] not in name2label:
            continue
        label = name2label[bbox['name']]
        x_center = (bbox['xmin']+bbox['xmax'])/(2*W)
        y_center = (bbox['ymin']+bbox['ymax'])/(2*H)
        width = (bbox['xmax']-bbox['xmin'])/W
        height = (bbox['ymax']-bbox['ymin'])/H
        row.append(
            ' '.join(list(map(str, [label, x_center, y_center, width, height]))))

    if not os.path.isdir(path_out):
        os.mkdir(path_out)
    file_name = os.path.splitext(os.path.basename(path_anno))[0]
    path_out_file = os.path.join(path_out, '{}.txt'.format(file_name))
    with open(path_out_file, 'w') as f:
        f.write('\n'.join(row))


def annos2labels(path_anno_folder, path_names, path_out):
    """
    Generate label txt from a list of annotations.

    Args:
        path_anno_folder: Path of the annotation files folder.
        path_names: Path of the .names file.
        path_out: Path of the output .txt file.

    Returns:
        None
    """
    path_annos = [os.path.join(path_anno_folder, i)
                  for i in os.listdir(path_anno_folder)]
    path_names_ = [path_names for i in range(len(path_annos))]
    path_out_ = [path_out for i in range(len(path_annos))]
    pool = Pool(cpu_count())
    pool.starmap(anno2label, zip(path_annos, path_names_,
                                 path_out_,))
    pool.close()
    pool.join()


def bbox_dist(path_anno_folder, verbose=0):
    """
    Analysis the bbox distribution by a list of annotation files.

    Args:
        path_anno_folder: Path of the annotation files folder.

    Returns：
        A dict contains the result.
    """
    annos = parse_annos(path_anno_folder)
    d = {}
    d['LEN'] = 0
    for annoname in annos:
        anno = annos[annoname]
        objs = anno['object']
        for obj in objs:
            d[obj['name']] = d.setdefault(obj['name'], 0)+1
            d['LEN'] += 1
    if verbose:
        for name in d:
            print('{}: {}({:.4f}%)'.format(
                name, d[name], 100*d[name]/d['LEN']))
    return d


def get_label(path_names):
    """
    Get name2label and label2name from a .names file.

    Args:
        path_names: Path of the .names file.

    Returns:
        name2label: A dict like {'name1': 0, 'name2': 1, ...}
        label2name: A dict like {0: 'name1', 1: 'name2', ...}
    """
    name2label = {}
    label2name = {}
    with open(path_names) as f:
        label = 0
        for l in f:
            name = l.strip('\n')
            name2label[name] = label
            label2name[label] = name
            label += 1
    return name2label, label2name


def base2abs(path_base, path_prefix, path_out):
    """
    Convert the base names to absolute paths.

    Args:
        path_base: Path of the base file.
        path_prefix: Prefix path you wanna put behind the base names.
        path_out: Path of the output file.

    Returns:
        None
    """
    prefix_abs = os.path.abspath(path_prefix)
    bases = []
    with open(path_base) as f:
        for line in f:
            bases.append(line.strip('\n'))

    bases_abs = [os.path.join(prefix_abs, i) for i in bases]
    with open(path_out, 'w') as f:
        f.write('\n'.join(bases_abs))


def split_trainval(path_img_folder, path_train, path_val, ratio_train):
    '''
    Split the date into training & validation subset by basename format.

    Param:
        path_img_folder: Path of JPEImages.
        path_train: Path of the txt file containing training subset.
        path_val: Path of the txt file containing validation subset.
        ratio_train: The ratio of training set which should be more than 0 and less or equal than 1.

    Return:
        None
    '''
    if not 0 < ratio_train <= 1:
        print('Please set a right ratio_train.')
        return
    name_all = os.listdir(path_img_folder)
    np.random.seed(0)
    np.random.shuffle(name_all)
    len_train = int(len(name_all)*ratio_train)
    name_train = name_all[:len_train]
    name_test = name_all[len_train:]
    with open(path_train, 'w') as f:
        f.write('\n'.join(name_train))
    with open(path_val, 'w') as f:
        f.write('\n'.join(name_test))
